(*****************************************************************************)
(*  graphic_server.ml  :                                                     *)
(*                                                                           *)
(* POD : E.Chailloux  P.Manoury  B.Pagano                                    *)
(*****************************************************************************)


open Graphics 
let si = string_of_int 


                    (* Les requetes graphiques *)

class virtual graph_query = 
 object
  method virtual draw : unit
end

class gq_clear =
 object
  inherit graph_query 
  method draw = clear_graph ()
end

class gq_line col (x1,y1) (x2,y2) = 
 object
  inherit graph_query 
  method draw = set_color col; moveto x1 y1; lineto x2 y2 ;
end

class gq_disk col (x,y) r = 
 object
  inherit graph_query 
  method draw = set_color col; fill_circle x y r;
end

(* la classe gq_kill est utiliser pour fermer le serveur *)
class gq_kill () =
 object
  inherit graph_query 
  method draw = failwith "gq_kill#draw : not an usable method"  
end

(* On definit des "constructeurs" pour chaque sous-classe de graph_query *)

let clear = new gq_clear
and line = new gq_line 
and disk = new gq_disk 







                      (* Le serveur graphique *)

open Event 

class graph_server x y = 
  object (self)
    (* canal sur lequel transitent les requetes graphiques *)
    val sync_channel = (new_channel () : graph_query channel) 

    (* samaphore pour forcer le traitement sequentiel des requetes*)
    val sem = Mutex.create ()

    (* requete qui servira a terminer le serveur *)
    val kill_signal = (new gq_kill () : graph_query)

    (* flag indiquant si le serveur est en etat d'accepter une requete *)
    val mutable flag_alive = false

    initializer
      (* boucle de reception et de traitement des requetes *)
      let boucle_receive_treat () =
	 while true do 
	   let gq = sync (receive sync_channel) 
           (* si la requete est kill on termine la boucle sinon on traite *)
	   in if gq==kill_signal then Thread.exit () 
                                else self#treat_query gq  
	 done
      in        
	open_graph (" " ^ (si x) ^ "x" ^ (si y)) ;
        (* la boucle s'execute dans un autre processus *)
	Thread.create boucle_receive_treat () ; 
	(* le serveur est operationnel *)
	flag_alive <- true 

    method get_channel = sync_channel
    method alive = flag_alive 

    method private treat_query gq=
      (* les traitements effectifs des requetes s'effectuent 
         un apres l'autre *)
      Mutex.lock sem; gq#draw; Mutex.unlock sem

    method exit = 
      (* le seurveur est considere comme arreter *)
      flag_alive <- false ;
      (* on envoie le signal kill pour terminer la boucle de reception *)
      sync (send sync_channel kill_signal) ;
      close_graph () 
end



	            (* le client graphique *)

class graph_client (c:#graph_server)  =
  object (self)
    method send gq = 
      if c#alive then sync (send c#get_channel gq) 
end












class enhanced_graph_server x y  =
  object (self)
    inherit graph_server x y

    (* un second canal pour les requetes asynchrones *)
    val async_channel = new_channel ()

    (* une pile FIFO pour stocker les requetes recues et non traitees *)
    val mutable query_fifo = []
    (* le traitement de cette pile est une section critique protegee *)
    val async_sem = Mutex.create ()
    val async_signal = Condition.create ()

    initializer 
      (* boucle de reception des requetes asynchrones *)
      let boucle_receive () =
	 while true do 
	   let gq = sync (receive async_channel) 
           in
	     begin
	       (* empiler la requete : section critique*)
	       Mutex.lock async_sem ;
	       query_fifo <- query_fifo @ [gq] ; 
	       Mutex.unlock async_sem ;
	       (* on reveille la boucle de traitement *)
	       Condition.signal async_signal ;
             (* si la requete est kill on termine la boucle sinon on traite *)
             (* NB: la requete kill est empilee pour terminer l'autre boucle *)
	       if gq==kill_signal then Thread.exit ()  
	     end 
	 done

      (* boucle de traitement des requetes asynchrones*)
      and boucle_treat () = 
	while true do
	   Mutex.lock async_sem ;
	   (* si la boucle est vide on libere le semaphore et on attend *)
	   (* le signal de reception d'une requete *)
 	   while query_fifo = [] do 
	     Condition.wait async_signal async_sem 
	   done ;
           (* si la requete est kill on termine sinon on traite *)
           let gq = List.hd query_fifo 
	   in if gq == kill_signal 
	     then (Mutex.unlock async_sem ; Thread.exit () )
             else 
	       self#treat_query gq  ;
	       query_fifo <- List.tl query_fifo ;
	       Mutex.unlock async_sem
	done

      in 
	Thread.create boucle_receive ();
	Thread.create boucle_treat ();	()
	
    method get_async_channel = async_channel
    method exit = 
      flag_alive <- false ;
      (* on envoie le signal kill sur les deux canaux *)
      sync (send sync_channel kill_signal) ;
      sync (send async_channel kill_signal) ;      
      close_graph () 
end 


class enhanced_graph_client (egs:#enhanced_graph_server) =
  object 
    inherit graph_client egs
    method async_send gq = 
      if egs#alive then sync (send egs#get_async_channel gq) 
end




