(*****************************************************************************)
(*  robot_def.ml  :                                                          *)
(*                                                                           *)
(* POD : E.Chailloux  P.Manoury  B.Pagano                                    *)
(*****************************************************************************)

(* prend un predicat et une liste et rend la sous-liste des *)
(* elements satisfaisant le predicat.                       *)
let rec select pred = function [] -> []
  | a::l -> let next = select pred l in if pred a then a::next else next

(* extraire au hasard un element d'une liste *)
let random_list = function 
    [] ->  failwith "random_list: bad argument"
  | l -> List.nth l (Random.int (List.length l))

let cases_voisines (x,y) = 
  [ (x-1,y-1);(x-1,y);(x-1,y+1);(x,y-1);(x,y+1);(x+1,y-1);(x+1,y);(x+1,y+1) ]



(*************************************************************************)
(*                       ROBOT ET MONDE VIRTUEL                          *)
(*************************************************************************)


(* la classe monde est parametree par le type de ses robots *)

class virtual ['a] mondeV hi li =
 object (s)
    (* un robot doit posseder les methodes get_pos et set_pos *)
    constraint 'a = <get_pos:int*int; set_pos:int*int -> unit; .. >

    val h = (hi:int)
    val l = (li:int)
    val mutable roblist = ([]:'a list)

    method virtual is_legal : (int * int) -> bool 
    method virtual normalize : (int * int) -> int * int 

    method get_roblist = roblist
    method get_dim = (h,l)

    method enter r = roblist <- r::roblist
    method exit r = roblist <- select (fun ro -> r<>ro) roblist 

    method is_free p = 
      List.for_all (fun r -> p <> s#normalize(r#get_pos)) roblist 

    method free_places pos =
      let f p = let p1=s#normalize p in (s#is_free p1) && (s#is_legal p1)
      in select f (cases_voisines pos)
end



class virtual ['a] robotV (xi,yi) (vxi, vyi)= 
  object
    (* un monde doit posseder la methode get_roblist *)
    constraint 'a = <get_roblist: 'a #robotV list ; ..>
    val mutable x = (xi:int)
    val mutable y = (yi:int)
    val mutable vx = (vxi:int)
    val mutable vy = (vyi:int)

    method virtual next_position :'a -> (int * int)

    method get_pos = (x,y)
    method get_speed = (vx,vy)
    method set_pos (nx,ny) = x<-nx ;y<-ny
    method set_speed (nvx,nvy) = vx<-nvx;vy<-nvy

  end 







(*************************************************************************)
(*                              MONDES CONCRETS                          *)
(*************************************************************************)

class ['a] monde_ferme h l = 
  object
    inherit ['a] mondeV h l
    method is_legal (a,b) = (a >= 0) && (a < l) && (b >= 0) && (b < h)
    method normalize p = p
end


class ['a] monde_rond h l =
  object
    inherit ['a] mondeV h l
    method is_legal _ = true 
    method normalize (x,y) = ( x mod l , y mod h )
end




(*************************************************************************)
(*                              ROBOTS CONCRETS                          *)
(*************************************************************************)



class virtual ['a] robot_commandV p v =
  object (self)
    inherit ['a] robotV p v 
      
    method private go_right_45 = 
      if vx=0 then vx <- vy 
      else if vy=0 then vy <- (-vx)
      else if vx*vy>0 then vy <- 0
      else vx <- 0

    method private go_left_45 = 
      if vx=0 then vx <- (-vy) 
      else if vy=0 then vy <- vx
      else if vx*vy>0 then vx <- 0
      else vy <- 0

    method private go_right_90 = self#go_right_45 ; self#go_right_45
    method private go_left_90  = self#go_left_45  ; self#go_left_45 
  end



                         (* Robot fixe *)

class ['a] robot_fixe p =
  object
    inherit ['a] robotV p (0,0)

    method next_position p = (x,y)
  end



                         (* Robots fous *)

class ['a] robot_fou p =
  object (s)
    inherit ['a] robotV p (0,0)

    method next_position m = 
      let list_pos = m#free_places  s#get_pos
      in if list_pos = [] then s#get_pos else random_list list_pos
  end


class ['a] robot_fou_inertie p v =
  object (s)
    inherit ['a] robot_commandV p v
    method next_position m = 
      (  match Random.int 3 with 
	     0 -> () 
	   | 1 -> s#go_left_45 
	   | _ -> s#go_right_45   );
      let npos = (x+vx,y+vy) 
      in if m#is_free npos && m#is_legal npos then npos else (x,y)
  end



                         (* Robot poli *)

class ['a] robot_poli p v = 
  object (s)
  inherit ['a] robot_commandV p v 

    method next_position m =
      let rob_pos_list = List.map (fun r -> r#get_pos) m#get_roblist
      in let voisin = 
        List.exists (fun p -> List.mem p rob_pos_list) (cases_voisines (x,y))
      in if voisin then (x,y)
         else let new_pos = (x+vx,y+vy)
	 in if m#is_legal new_pos && m#is_free new_pos then new_pos 
            else ( s#go_right_90 ; (x,y) )
  end


                         (* Robot presse *)

class ['a] robot_presse p v =
  object (s)
  inherit ['a] robot_commandV p v 

    method next_position m =
      let new_pos = (x+vx,y+vy) 
      in if m#is_legal new_pos && m#is_free new_pos then new_pos 
         else ( s#go_left_45 ; (x,y) )
  end



                         (* Robot amical *)

class ['a] robot_amical p v (a:'a #robotV) = 
  object (s)
    inherit ['a] robot_commandV p v 
    val mutable ami = a 

    method private est_eloigne r = 
      let (rx,ry) = r#get_pos in (abs (x-rx)) > 1  ||  (abs (y-ry)) > 1

    method private nouvel_ami m = 
      let liste_possible = select s#est_eloigne m#get_roblist
      in if liste_possible = [] then s 
	                        else random_list liste_possible

    method private change_ami m = ami <- s#nouvel_ami m

    method next_position m = 
      (if not (s#est_eloigne ami) then s#change_ami m );
      let (ax,ay)=ami#get_pos 
      in let sign n = if n=0 then 0 else n / (abs n)
      in vx <- sign (ax-x) ; vy <- sign (ay-y) ;	( x+vx , y+vy ) 
end



let f() = 
 let m1 = new monde_ferme 10 10 
 and r1 = new robot_poli (1,1) (1,1) 
 in let r2 =  new robot_amical (1,1) (1,1) r1 
 in m1#enter r1; m1#enter r2 ;;


